from __future__ import annotations

from .cowsay import cowsay_art_definition, register_cowsay_art, render_cowsay

__all__ = [
    "render_cowsay",
    "cowsay_art_definition",
    "register_cowsay_art",
]
